import os, sys, struct, zlib, random

tiledatafilename = sys.argv[1]
outputfilename = sys.argv[2]

map_version = 2
endian = "!" # network (big)
endian = "@" # native

type_spawnpoint = 0x0001

type_image = 0x8001
type_tilemap = 0x8002
type_collision = 0x8003

data = ""
items = ""
offsets = []

class c_item_type:
	def __init__(self):
		self.type = 0
		self.start = 0
		self.num = 0

class c_map_file:
	def __init__(self):
		self.data = ""
		self.items = ""
		self.num_datas = 0
		self.num_items = 0
		self.data_offsets = []
		self.item_offsets = []
		self.item_types = []
		self.current_type = -1
		
	def add_item(self, type, id, item):
		#print "item: %x %x %x" % (type, id, len(item))
		
		if type < self.current_type:
			print "Type error!!!"

		if type == self.current_type:
			self.item_types[-1].num += 1
		else:
			t = c_item_type()
			t.type = type
			t.start = len(self.item_offsets)
			t.num = 1
			self.item_types += [t]
		
		self.item_offsets += [len(self.items)]
		self.items += struct.pack(endian + "II", type*(256*256)+id, len(item)) + item
		index = self.num_items
		self.num_items += 1
		self.current_type = type
		return index
	
	def add_data(self, data):
		#print "data: %x" % len(data)
		
		self.data_offsets += [len(self.data)]
		self.data += data
		index = self.num_datas
		self.num_datas += 1
		return index
	
	def write(self, filename):
		offsets = ""
		types = ""
		for i in self.item_types:
			types += struct.pack(endian + "iii", i.type, i.start, i.num)
		for i in self.item_offsets:
			offsets += struct.pack(endian + "i", i)
		for i in self.data_offsets:
			offsets += struct.pack(endian + "i", i)
		
		# attach header
		out = struct.pack(endian + "iiiii",
				len(self.item_types), # types
				self.num_items,
				self.num_datas,
				len(self.items),
				len(self.data))
		
		# add types, offsets and items
		out += types
		out += offsets
		out += self.items
				
		# get swapsize
		swaplen = len(out)
		
		# append raw data
		out += self.data

		# wrap it up
		out = struct.pack("!4siii", "MAPP", map_version, len(out), swaplen) + out
		#print "finish: %x" % (len(out))
		file(filename, "wb").write(out)
		#print self.item_offsets, self.data_offsets

mapfile = c_map_file()


spawnpoints = []
valid = [8,16,32,64,128,256,512,1024,1024*2,1024*4]

lines = file(tiledatafilename).readlines()
image = file("data_src/tilemap.tga", "rb").read()
img_type = struct.unpack("<B", image[2:3])[0]
img_bpp = struct.unpack("<B", image[16:17])[0]
img_width = struct.unpack("<H", image[12:14])[0]
img_height = struct.unpack("<H", image[14:16])[0]

if img_type != 2 or img_bpp != 32:
	print "image must be a RGBA"
	sys.exit(-1)
if not img_width in valid:
	print "width of the image is not valid"
	sys.exit(-1)
if not img_height in valid:
	print "height of the image is not valid"
	sys.exit(-1)

start = 18+struct.unpack("B", image[0])[0]
end = start + img_width*img_height*4
image_data = image[start:end] # fetch raw data

#print "Image:", img_width, img_height, len(image_data), img_width*img_height*4

width = 0
height = len(lines)
images = 1
tilemaps = 1
for l in lines:
	n = len(l.strip())
	if n > width:
		width = n

# tilemap header
# image, width, height, x, y, scale(fixed point)
tilemap_data = []
y = 0
for l in lines:
	x = 0
	for c in l.strip():
		if c == '#':
			tilemap_data += [1]
		else:
			tilemap_data += [0]
		
		if c == 's':
			spawnpoints += [[x*32+16,y*32+16]]
		x += 1
			
	for c in range(len(l), width-1):
		tilemap_data += [0]
	y += 1
		
if len(tilemap_data) != width*height:
	print "something is fishy with the tilemap data"

class c_tilemap_fixer:
	def __init__(self, data, w, h):
		self.data = data
		self.w = w
		self.h = h
	
	def get(self,x,y):
		if x < 0 or x >= self.w or y >= self.h:
			return 1 # solid
		if y < 0:
			return 0 # air
		if self.data[y*self.w+x] != 0:
			return 1
		return 0
		
	def set(self,x,y,u,v):
		self.data[y*self.w+x] = v*16+u
	
	def nice(self):
		for x in range(0, self.w):
			for y in range(0, self.h):
				if self.get(x,y): # solid
					solids = self.get(x+1,y)
					solids += self.get(x-1,y)
					solids += self.get(x,y+1)
					solids += self.get(x,y-1)

					corners = self.get(x+1,y+1)
					corners += self.get(x-1,y+1)
					corners += self.get(x+1,y-1)
					corners += self.get(x-1,y-1)
					
					if solids == 4 and corners == 3:
						if not self.get(x+1,y+1): self.set(x,y,2,3)
						elif not self.get(x-1,y+1): self.set(x,y,3,3)
						elif not self.get(x+1,y-1): self.set(x,y,1,3)
						elif not self.get(x-1,y-1): self.set(x,y,0,3)
					if solids == 3:
						if not self.get(x,y-1): self.set(x,y,0,1)
						elif not self.get(x,y+1): self.set(x,y,2,1)
						elif not self.get(x-1,y): self.set(x,y,3,1)
						elif not self.get(x+1,y): self.set(x,y,1,1)
					elif solids == 2:
						if not self.get(x,y-1) and not self.get(x-1,y): self.set(x,y,0,2)
						elif not self.get(x,y-1) and not self.get(x+1,y): self.set(x,y,1,2)
						elif not self.get(x,y+1) and not self.get(x+1,y): self.set(x,y,2,2)
						elif not self.get(x,y+1) and not self.get(x-1,y): self.set(x,y,3,2)
						if not self.get(x,y-1) and not self.get(x-1,y): self.set(x,y,0,2)
					elif corners == 4 and solids == 4:
						if random.randint(0, 10) == 0:
							self.set(x,y,2+random.choice(range(0,2)),0)
							
	
	def tilemap_data(self):
		d = ""
		for i in self.data:
			d += struct.pack("!B", i)
		return d

	def collision_data(self):
		d = ""
		for i in self.data:
			if i == 0:
				d += struct.pack("!B", 0)
			else:
				d += struct.pack("!B", 1)
		return d
					

tilemap_fixer = c_tilemap_fixer(tilemap_data, width, height)
tilemap_fixer.nice()
tilemap_data = tilemap_fixer.tilemap_data()
collision_data = tilemap_fixer.collision_data()

# add spawnpoints
id = 0
for p in spawnpoints:
	sp_item = struct.pack(endian+"ii", p[0], p[1])
	mapfile.add_item(type_spawnpoint, id, sp_item)
	id += 1

# add image
image_item = struct.pack(endian+"iii", img_width, img_height, mapfile.add_data(image_data))
image_id = mapfile.add_item(type_image, 0, image_item)

# add tilemap
tilemap_item = struct.pack(endian+"iiiiiii", 0, width, height, 0, 0, 1*(256*256), mapfile.add_data(tilemap_data))
tilemap_id = mapfile.add_item(type_tilemap, 0, tilemap_item)

# add collision
collision_item = struct.pack(endian+"iii", width, height, mapfile.add_data(collision_data))
collision_id = mapfile.add_item(type_collision, 0, collision_item)

mapfile.write(outputfilename)


